@@ -16,10 +16,7 @@ angular.module('syncthing.core')
                             }
                         });
                         //Prevents user from adding a duplicate ID
-                        var matches = scope.devices.filter(function (n) {
-                            return n.deviceID == viewValue;
-                        }).length;
-                        if (matches > 0) {
+                        if (scope.devices.hasOwnProperty(viewValue)) {
                             ctrl.$setValidity('unique', false);
                         } else {
                             ctrl.$setValidity('unique', true);
