@@ -17,14 +17,13 @@ import (
 
 	"github.com/syncthing/syncthing/lib/config"
 	"github.com/syncthing/syncthing/lib/protocol"
-	"github.com/syncthing/syncthing/lib/sync"
+	"github.com/syncthing/syncthing/lib/util"
 )
 
 type configMuxBuilder struct {
 	*httprouter.Router
 	id  protocol.DeviceID
 	cfg config.Wrapper
-	mut sync.Mutex
 }
 
 func (c *configMuxBuilder) registerConfig(path string) {
@@ -53,20 +52,31 @@ func (c *configMuxBuilder) registerConfigInsync(path string) {
 	})
 }
 
+func (c *configMuxBuilder) registerConfigRequiresRestart(path string) {
+	c.HandlerFunc(http.MethodGet, path, func(w http.ResponseWriter, _ *http.Request) {
+		sendJSON(w, map[string]bool{"requiresRestart": c.cfg.RequiresRestart()})
+	})
+}
+
 func (c *configMuxBuilder) registerFolders(path string) {
 	c.HandlerFunc(http.MethodGet, path, func(w http.ResponseWriter, _ *http.Request) {
 		sendJSON(w, c.cfg.FolderList())
 	})
 
 	c.HandlerFunc(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request) {
-		c.mut.Lock()
-		defer c.mut.Unlock()
-		var folders []config.FolderConfiguration
-		if err := unmarshalTo(r.Body, &folders); err != nil {
-			http.Error(w, err.Error(), http.StatusBadRequest)
-			return
+		data, err := unmarshalToRawMessages(r.Body)
+		folders := make([]config.FolderConfiguration, len(data))
+		defaultFolder := c.cfg.DefaultFolder()
+		for i, bs := range data {
+			folders[i] = defaultFolder.Copy()
+			if err := json.Unmarshal(bs, &folders[i]); err != nil {
+				http.Error(w, err.Error(), http.StatusBadRequest)
+				return
+			}
 		}
-		waiter, err := c.cfg.SetFolders(folders)
+		waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+			cfg.SetFolders(folders)
+		})
 		if err != nil {
 			http.Error(w, err.Error(), http.StatusInternalServerError)
 			return
@@ -75,19 +85,7 @@ func (c *configMuxBuilder) registerFolders(path string) {
 	})
 
 	c.HandlerFunc(http.MethodPost, path, func(w http.ResponseWriter, r *http.Request) {
-		c.mut.Lock()
-		defer c.mut.Unlock()
-		var folder config.FolderConfiguration
-		if err := unmarshalTo(r.Body, &folder); err != nil {
-			http.Error(w, err.Error(), http.StatusBadRequest)
-			return
-		}
-		waiter, err := c.cfg.SetFolder(folder)
-		if err != nil {
-			http.Error(w, err.Error(), http.StatusInternalServerError)
-			return
-		}
-		c.finish(w, waiter)
+		c.adjustFolder(w, r, c.cfg.DefaultFolder(), false)
 	})
 }
 
@@ -97,14 +95,19 @@ func (c *configMuxBuilder) registerDevices(path string) {
 	})
 
 	c.HandlerFunc(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request) {
-		c.mut.Lock()
-		defer c.mut.Unlock()
-		var devices []config.DeviceConfiguration
-		if err := unmarshalTo(r.Body, &devices); err != nil {
-			http.Error(w, err.Error(), http.StatusBadRequest)
-			return
+		data, err := unmarshalToRawMessages(r.Body)
+		devices := make([]config.DeviceConfiguration, len(data))
+		defaultDevice := c.cfg.DefaultDevice()
+		for i, bs := range data {
+			devices[i] = defaultDevice.Copy()
+			if err := json.Unmarshal(bs, &devices[i]); err != nil {
+				http.Error(w, err.Error(), http.StatusBadRequest)
+				return
+			}
 		}
-		waiter, err := c.cfg.SetDevices(devices)
+		waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+			cfg.SetDevices(devices)
+		})
 		if err != nil {
 			http.Error(w, err.Error(), http.StatusInternalServerError)
 			return
@@ -113,19 +116,7 @@ func (c *configMuxBuilder) registerDevices(path string) {
 	})
 
 	c.HandlerFunc(http.MethodPost, path, func(w http.ResponseWriter, r *http.Request) {
-		c.mut.Lock()
-		defer c.mut.Unlock()
-		var device config.DeviceConfiguration
-		if err := unmarshalTo(r.Body, &device); err != nil {
-			http.Error(w, err.Error(), http.StatusBadRequest)
-			return
-		}
-		waiter, err := c.cfg.SetDevice(device)
-		if err != nil {
-			http.Error(w, err.Error(), http.StatusInternalServerError)
-			return
-		}
-		c.finish(w, waiter)
+		c.adjustDevice(w, r, c.cfg.DefaultDevice(), false)
 	})
 }
 
@@ -140,7 +131,7 @@ func (c *configMuxBuilder) registerFolder(path string) {
 	})
 
 	c.Handle(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request, p httprouter.Params) {
-		c.adjustFolder(w, r, config.FolderConfiguration{})
+		c.adjustFolder(w, r, c.cfg.DefaultFolder(), false)
 	})
 
 	c.Handle(http.MethodPatch, path, func(w http.ResponseWriter, r *http.Request, p httprouter.Params) {
@@ -149,12 +140,10 @@ func (c *configMuxBuilder) registerFolder(path string) {
 			http.Error(w, "No folder with given ID", http.StatusNotFound)
 			return
 		}
-		c.adjustFolder(w, r, folder)
+		c.adjustFolder(w, r, folder, false)
 	})
 
 	c.Handle(http.MethodDelete, path, func(w http.ResponseWriter, _ *http.Request, p httprouter.Params) {
-		c.mut.Lock()
-		defer c.mut.Unlock()
 		waiter, err := c.cfg.RemoveFolder(p.ByName("id"))
 		if err != nil {
 			http.Error(w, err.Error(), http.StatusBadRequest)
@@ -186,19 +175,21 @@ func (c *configMuxBuilder) registerDevice(path string) {
 	})
 
 	c.Handle(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request, p httprouter.Params) {
-		c.adjustDevice(w, r, config.DeviceConfiguration{})
+		c.adjustDevice(w, r, c.cfg.DefaultDevice(), false)
 	})
 
 	c.Handle(http.MethodPatch, path, func(w http.ResponseWriter, r *http.Request, p httprouter.Params) {
 		if device, ok := deviceFromParams(w, p); ok {
-			c.adjustDevice(w, r, device)
+			c.adjustDevice(w, r, device, false)
 		}
 	})
 
 	c.Handle(http.MethodDelete, path, func(w http.ResponseWriter, _ *http.Request, p httprouter.Params) {
-		c.mut.Lock()
-		defer c.mut.Unlock()
 		id, err := protocol.DeviceIDFromString(p.ByName("id"))
+		if err != nil {
+			http.Error(w, err.Error(), http.StatusBadRequest)
+			return
+		}
 		waiter, err := c.cfg.RemoveDevice(id)
 		if err != nil {
 			http.Error(w, err.Error(), http.StatusBadRequest)
@@ -208,13 +199,47 @@ func (c *configMuxBuilder) registerDevice(path string) {
 	})
 }
 
+func (c *configMuxBuilder) registerDefaultFolder(path string) {
+	c.HandlerFunc(http.MethodGet, path, func(w http.ResponseWriter, _ *http.Request) {
+		sendJSON(w, c.cfg.DefaultFolder())
+	})
+
+	c.HandlerFunc(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request) {
+		var cfg config.FolderConfiguration
+		util.SetDefaults(&cfg)
+		c.adjustFolder(w, r, cfg, true)
+	})
+
+	c.HandlerFunc(http.MethodPatch, path, func(w http.ResponseWriter, r *http.Request) {
+		c.adjustFolder(w, r, c.cfg.DefaultFolder(), true)
+	})
+}
+
+func (c *configMuxBuilder) registerDefaultDevice(path string) {
+	c.HandlerFunc(http.MethodGet, path, func(w http.ResponseWriter, _ *http.Request) {
+		sendJSON(w, c.cfg.DefaultDevice())
+	})
+
+	c.HandlerFunc(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request) {
+		var cfg config.DeviceConfiguration
+		util.SetDefaults(&cfg)
+		c.adjustDevice(w, r, cfg, true)
+	})
+
+	c.HandlerFunc(http.MethodPatch, path, func(w http.ResponseWriter, r *http.Request) {
+		c.adjustDevice(w, r, c.cfg.DefaultDevice(), true)
+	})
+}
+
 func (c *configMuxBuilder) registerOptions(path string) {
 	c.HandlerFunc(http.MethodGet, path, func(w http.ResponseWriter, _ *http.Request) {
 		sendJSON(w, c.cfg.Options())
 	})
 
 	c.HandlerFunc(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request) {
-		c.adjustOptions(w, r, config.OptionsConfiguration{})
+		var cfg config.OptionsConfiguration
+		util.SetDefaults(&cfg)
+		c.adjustOptions(w, r, cfg)
 	})
 
 	c.HandlerFunc(http.MethodPatch, path, func(w http.ResponseWriter, r *http.Request) {
@@ -228,7 +253,9 @@ func (c *configMuxBuilder) registerLDAP(path string) {
 	})
 
 	c.HandlerFunc(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request) {
-		c.adjustLDAP(w, r, config.LDAPConfiguration{})
+		var cfg config.LDAPConfiguration
+		util.SetDefaults(&cfg)
+		c.adjustLDAP(w, r, cfg)
 	})
 
 	c.HandlerFunc(http.MethodPatch, path, func(w http.ResponseWriter, r *http.Request) {
@@ -242,7 +269,9 @@ func (c *configMuxBuilder) registerGUI(path string) {
 	})
 
 	c.HandlerFunc(http.MethodPut, path, func(w http.ResponseWriter, r *http.Request) {
-		c.adjustGUI(w, r, config.GUIConfiguration{})
+		var cfg config.GUIConfiguration
+		util.SetDefaults(&cfg)
+		c.adjustGUI(w, r, cfg)
 	})
 
 	c.HandlerFunc(http.MethodPatch, path, func(w http.ResponseWriter, r *http.Request) {
@@ -251,36 +280,45 @@ func (c *configMuxBuilder) registerGUI(path string) {
 }
 
 func (c *configMuxBuilder) adjustConfig(w http.ResponseWriter, r *http.Request) {
-	c.mut.Lock()
-	defer c.mut.Unlock()
-	cfg, err := config.ReadJSON(r.Body, c.id)
+	to, err := config.ReadJSON(r.Body, c.id)
 	r.Body.Close()
 	if err != nil {
 		l.Warnln("Decoding posted config:", err)
 		http.Error(w, err.Error(), http.StatusBadRequest)
 		return
 	}
-	if cfg.GUI.Password, err = checkGUIPassword(c.cfg.GUI().Password, cfg.GUI.Password); err != nil {
-		l.Warnln("bcrypting password:", err)
-		http.Error(w, err.Error(), http.StatusInternalServerError)
-		return
-	}
-	waiter, err := c.cfg.Replace(cfg)
-	if err != nil {
+	var errMsg string
+	var status int
+	waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+		if to.GUI.Password, err = checkGUIPassword(cfg.GUI.Password, to.GUI.Password); err != nil {
+			l.Warnln("bcrypting password:", err)
+			errMsg = err.Error()
+			status = http.StatusInternalServerError
+			return
+		}
+		*cfg = to
+	})
+	if errMsg != "" {
+		http.Error(w, errMsg, status)
+	} else if err != nil {
 		http.Error(w, err.Error(), http.StatusInternalServerError)
 		return
 	}
 	c.finish(w, waiter)
 }
 
-func (c *configMuxBuilder) adjustFolder(w http.ResponseWriter, r *http.Request, folder config.FolderConfiguration) {
-	c.mut.Lock()
-	defer c.mut.Unlock()
+func (c *configMuxBuilder) adjustFolder(w http.ResponseWriter, r *http.Request, folder config.FolderConfiguration, defaults bool) {
 	if err := unmarshalTo(r.Body, &folder); err != nil {
 		http.Error(w, err.Error(), http.StatusBadRequest)
 		return
 	}
-	waiter, err := c.cfg.SetFolder(folder)
+	waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+		if defaults {
+			cfg.Defaults.Folder = folder
+		} else {
+			cfg.SetFolder(folder)
+		}
+	})
 	if err != nil {
 		http.Error(w, err.Error(), http.StatusInternalServerError)
 		return
@@ -288,14 +326,18 @@ func (c *configMuxBuilder) adjustFolder(w http.ResponseWriter, r *http.Request,
 	c.finish(w, waiter)
 }
 
-func (c *configMuxBuilder) adjustDevice(w http.ResponseWriter, r *http.Request, device config.DeviceConfiguration) {
-	c.mut.Lock()
-	defer c.mut.Unlock()
+func (c *configMuxBuilder) adjustDevice(w http.ResponseWriter, r *http.Request, device config.DeviceConfiguration, defaults bool) {
 	if err := unmarshalTo(r.Body, &device); err != nil {
 		http.Error(w, err.Error(), http.StatusBadRequest)
 		return
 	}
-	waiter, err := c.cfg.SetDevice(device)
+	waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+		if defaults {
+			cfg.Defaults.Device = device
+		} else {
+			cfg.SetDevice(device)
+		}
+	})
 	if err != nil {
 		http.Error(w, err.Error(), http.StatusInternalServerError)
 		return
@@ -304,13 +346,13 @@ func (c *configMuxBuilder) adjustDevice(w http.ResponseWriter, r *http.Request,
 }
 
 func (c *configMuxBuilder) adjustOptions(w http.ResponseWriter, r *http.Request, opts config.OptionsConfiguration) {
-	c.mut.Lock()
-	defer c.mut.Unlock()
 	if err := unmarshalTo(r.Body, &opts); err != nil {
 		http.Error(w, err.Error(), http.StatusBadRequest)
 		return
 	}
-	waiter, err := c.cfg.SetOptions(opts)
+	waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+		cfg.Options = opts
+	})
 	if err != nil {
 		http.Error(w, err.Error(), http.StatusInternalServerError)
 		return
@@ -319,21 +361,26 @@ func (c *configMuxBuilder) adjustOptions(w http.ResponseWriter, r *http.Request,
 }
 
 func (c *configMuxBuilder) adjustGUI(w http.ResponseWriter, r *http.Request, gui config.GUIConfiguration) {
-	c.mut.Lock()
-	defer c.mut.Unlock()
 	oldPassword := gui.Password
 	err := unmarshalTo(r.Body, &gui)
 	if err != nil {
 		http.Error(w, err.Error(), http.StatusBadRequest)
 		return
 	}
-	if gui.Password, err = checkGUIPassword(oldPassword, gui.Password); err != nil {
-		l.Warnln("bcrypting password:", err)
-		http.Error(w, err.Error(), http.StatusInternalServerError)
-		return
-	}
-	waiter, err := c.cfg.SetGUI(gui)
-	if err != nil {
+	var errMsg string
+	var status int
+	waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+		if gui.Password, err = checkGUIPassword(oldPassword, gui.Password); err != nil {
+			l.Warnln("bcrypting password:", err)
+			errMsg = err.Error()
+			status = http.StatusInternalServerError
+			return
+		}
+		cfg.GUI = gui
+	})
+	if errMsg != "" {
+		http.Error(w, errMsg, status)
+	} else if err != nil {
 		http.Error(w, err.Error(), http.StatusInternalServerError)
 		return
 	}
@@ -341,13 +388,13 @@ func (c *configMuxBuilder) adjustGUI(w http.ResponseWriter, r *http.Request, gui
 }
 
 func (c *configMuxBuilder) adjustLDAP(w http.ResponseWriter, r *http.Request, ldap config.LDAPConfiguration) {
-	c.mut.Lock()
-	defer c.mut.Unlock()
 	if err := unmarshalTo(r.Body, &ldap); err != nil {
 		http.Error(w, err.Error(), http.StatusBadRequest)
 		return
 	}
-	waiter, err := c.cfg.SetLDAP(ldap)
+	waiter, err := c.cfg.Modify(func(cfg *config.Configuration) {
+		cfg.LDAP = ldap
+	})
 	if err != nil {
 		http.Error(w, err.Error(), http.StatusInternalServerError)
 		return
@@ -365,6 +412,12 @@ func unmarshalTo(body io.ReadCloser, to interface{}) error {
 	return json.Unmarshal(bs, to)
 }
 
+func unmarshalToRawMessages(body io.ReadCloser) ([]json.RawMessage, error) {
+	var data []json.RawMessage
+	err := unmarshalTo(body, &data)
+	return data, err
+}
+
 func checkGUIPassword(oldPassword, newPassword string) (string, error) {
 	if newPassword == oldPassword {
 		return newPassword, nil
