@@ -10,6 +10,8 @@ import (
 	"testing"
 
 	"github.com/syncthing/syncthing/lib/config"
+	"github.com/syncthing/syncthing/lib/events"
+	"github.com/syncthing/syncthing/lib/protocol"
 )
 
 func TestIsLANHost(t *testing.T) {
@@ -35,7 +37,7 @@ func TestIsLANHost(t *testing.T) {
 		Options: config.OptionsConfiguration{
 			AlwaysLocalNets: []string{"10.20.30.0/24"},
 		},
-	})
+	}, protocol.LocalDeviceID, events.NoopLogger)
 	s := &service{cfg: cfg}
 
 	for _, tc := range cases {
