@@ -7,6 +7,7 @@
 package events
 
 import (
+	"context"
 	"encoding/json"
 	"fmt"
 	"sync"
@@ -27,25 +28,30 @@ func TestNewLogger(t *testing.T) {
 	}
 }
 
-func TestSubscriber(t *testing.T) {
+func setupLogger() (Logger, context.CancelFunc) {
+	ctx, cancel := context.WithCancel(context.Background())
 	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	go l.Serve(ctx)
+	return l, cancel
+}
+
+func TestSubscriber(t *testing.T) {
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(0)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	if s == nil {
 		t.Fatal("Unexpected nil Subscription")
 	}
 }
 
 func TestTimeout(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(0)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	_, err := s.Poll(timeout)
 	if err != ErrTimeout {
 		t.Fatal("Unexpected non-Timeout error:", err)
@@ -53,13 +59,12 @@ func TestTimeout(t *testing.T) {
 }
 
 func TestEventBeforeSubscribe(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	l.Log(DeviceConnected, "foo")
 	s := l.Subscribe(0)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 
 	_, err := s.Poll(timeout)
 	if err != ErrTimeout {
@@ -68,12 +73,11 @@ func TestEventBeforeSubscribe(t *testing.T) {
 }
 
 func TestEventAfterSubscribe(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(AllEvents)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	l.Log(DeviceConnected, "foo")
 
 	ev, err := s.Poll(timeout)
@@ -95,12 +99,11 @@ func TestEventAfterSubscribe(t *testing.T) {
 }
 
 func TestEventAfterSubscribeIgnoreMask(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(DeviceDisconnected)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	l.Log(DeviceConnected, "foo")
 
 	_, err := s.Poll(timeout)
@@ -110,12 +113,11 @@ func TestEventAfterSubscribeIgnoreMask(t *testing.T) {
 }
 
 func TestBufferOverflow(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(AllEvents)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 
 	// The first BufferSize events will be logged pretty much
 	// instantaneously. The next BufferSize events will each block for up to
@@ -135,9 +137,8 @@ func TestBufferOverflow(t *testing.T) {
 }
 
 func TestUnsubscribe(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(AllEvents)
 	l.Log(DeviceConnected, "foo")
@@ -147,7 +148,7 @@ func TestUnsubscribe(t *testing.T) {
 		t.Fatal("Unexpected error:", err)
 	}
 
-	l.Unsubscribe(s)
+	s.Unsubscribe()
 	l.Log(DeviceConnected, "foo")
 
 	_, err = s.Poll(timeout)
@@ -157,12 +158,11 @@ func TestUnsubscribe(t *testing.T) {
 }
 
 func TestGlobalIDs(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(AllEvents)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	l.Log(DeviceConnected, "foo")
 	l.Subscribe(AllEvents)
 	l.Log(DeviceConnected, "bar")
@@ -189,12 +189,11 @@ func TestGlobalIDs(t *testing.T) {
 }
 
 func TestSubscriptionIDs(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(DeviceConnected)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 
 	l.Log(DeviceDisconnected, "a")
 	l.Log(DeviceConnected, "b")
@@ -231,12 +230,11 @@ func TestSubscriptionIDs(t *testing.T) {
 }
 
 func TestBufferedSub(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(AllEvents)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	bs := NewBufferedSubscription(s, 10*BufferSize)
 
 	go func() {
@@ -262,12 +260,11 @@ func TestBufferedSub(t *testing.T) {
 }
 
 func BenchmarkBufferedSub(b *testing.B) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(AllEvents)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	bufferSize := BufferSize
 	bs := NewBufferedSubscription(s, bufferSize)
 
@@ -318,12 +315,11 @@ func BenchmarkBufferedSub(b *testing.B) {
 }
 
 func TestSinceUsesSubscriptionId(t *testing.T) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(DeviceConnected)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	bs := NewBufferedSubscription(s, 10*BufferSize)
 
 	l.Log(DeviceConnected, "a") // SubscriptionID = 1
@@ -375,9 +371,8 @@ func TestUnsubscribeContention(t *testing.T) {
 		senders   = 1000
 	)
 
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	// Start listeners. These will poll until the stop channel is closed,
 	// then exit and unsubscribe.
@@ -390,7 +385,7 @@ func TestUnsubscribeContention(t *testing.T) {
 			defer listenerWg.Done()
 
 			s := l.Subscribe(AllEvents)
-			defer l.Unsubscribe(s)
+			defer s.Unsubscribe()
 
 			for {
 				select {
@@ -444,12 +439,11 @@ func TestUnsubscribeContention(t *testing.T) {
 }
 
 func BenchmarkLogEvent(b *testing.B) {
-	l := NewLogger()
-	defer l.Stop()
-	go l.Serve()
+	l, cancel := setupLogger()
+	defer cancel()
 
 	s := l.Subscribe(AllEvents)
-	defer l.Unsubscribe(s)
+	defer s.Unsubscribe()
 	NewBufferedSubscription(s, 1) // runs in the background
 
 	for i := 0; i < b.N; i++ {
