@@ -8,7 +8,10 @@
 
 package fs
 
-import "syscall"
+import (
+	"os"
+	"syscall"
+)
 
 func (e basicFileInfo) Mode() FileMode {
 	return FileMode(e.FileInfo.Mode())
@@ -27,3 +30,9 @@ func (e basicFileInfo) Group() int {
 	}
 	return -1
 }
+
+// fileStat converts e to os.FileInfo that is suitable
+// to be passed to os.SameFile. Non-trivial on Windows.
+func (e *basicFileInfo) osFileInfo() os.FileInfo {
+	return e.FileInfo
+}
