@@ -10,8 +10,9 @@ package fs
 
 import "github.com/syncthing/notify"
 
+// notify.InAttrib is not only required for permissions, but also mod. time changes
 const (
-	subEventMask  = notify.InCreate | notify.InMovedTo | notify.InDelete | notify.InDeleteSelf | notify.InModify | notify.InMovedFrom | notify.InMoveSelf
-	permEventMask = notify.InAttrib
+	subEventMask  = notify.InCreate | notify.InMovedTo | notify.InDelete | notify.InDeleteSelf | notify.InModify | notify.InMovedFrom | notify.InMoveSelf | notify.InAttrib
+	permEventMask = 0
 	rmEventMask   = notify.InDelete | notify.InDeleteSelf | notify.InMovedFrom | notify.InMoveSelf
 )
