@@ -27,7 +27,7 @@ import (
 // see readShortAt()
 const randomBlockShift = 14 // 128k
 
-// fakefs is a fake filesystem for testing and benchmarking. It has the
+// fakeFS is a fake filesystem for testing and benchmarking. It has the
 // following properties:
 //
 // - File metadata is kept in RAM. Specifically, we remember which files and
@@ -37,7 +37,7 @@ const randomBlockShift = 14 // 128k
 // - File contents are generated pseudorandomly with just the file name as
 //   seed. Writes are discarded, other than having the effect of increasing
 //   the file size. If you only write data that you've read from a file with
-//   the same name on a different fakefs, you'll never know the difference...
+//   the same name on a different fakeFS, you'll never know the difference...
 //
 // - We totally ignore permissions - pretend you are root.
 //
@@ -51,10 +51,10 @@ const randomBlockShift = 14 // 128k
 //     insens=b   "true" makes filesystem case-insensitive Windows- or OSX-style (default false)
 //     latency=d  to set the amount of time each "disk" operation takes, where d is time.ParseDuration format
 //
-// - Two fakefs:s pointing at the same root path see the same files.
+// - Two fakeFS:s pointing at the same root path see the same files.
 //
-type fakefs struct {
-	counters    fakefsCounters
+type fakeFS struct {
+	counters    fakeFSCounters
 	uri         string
 	mut         sync.Mutex
 	root        *fakeEntry
@@ -63,7 +63,7 @@ type fakefs struct {
 	latency     time.Duration
 }
 
-type fakefsCounters struct {
+type fakeFSCounters struct {
 	Chmod       int64
 	Lchown      int64
 	Chtimes     int64
@@ -81,13 +81,13 @@ type fakefsCounters struct {
 }
 
 var (
-	fakefsMut sync.Mutex
-	fakefsFs  = make(map[string]*fakefs)
+	fakeFSMut   sync.Mutex
+	fakeFSCache = make(map[string]*fakeFS)
 )
 
-func newFakeFilesystem(rootURI string, _ ...Option) *fakefs {
-	fakefsMut.Lock()
-	defer fakefsMut.Unlock()
+func newFakeFilesystem(rootURI string, _ ...Option) *fakeFS {
+	fakeFSMut.Lock()
+	defer fakeFSMut.Unlock()
 
 	root := rootURI
 	var params url.Values
@@ -97,12 +97,12 @@ func newFakeFilesystem(rootURI string, _ ...Option) *fakefs {
 		params = uri.Query()
 	}
 
-	if fs, ok := fakefsFs[rootURI]; ok {
+	if fs, ok := fakeFSCache[rootURI]; ok {
 		// Already have an fs at this path
 		return fs
 	}
 
-	fs := &fakefs{
+	fs := &fakeFS{
 		uri: "fake://" + rootURI,
 		root: &fakeEntry{
 			name:      "/",
@@ -157,7 +157,7 @@ func newFakeFilesystem(rootURI string, _ ...Option) *fakefs {
 	// the filesystem initially.
 	fs.latency, _ = time.ParseDuration(params.Get("latency"))
 
-	fakefsFs[root] = fs
+	fakeFSCache[root] = fs
 	return fs
 }
 
@@ -183,10 +183,10 @@ type fakeEntry struct {
 	content   []byte
 }
 
-func (fs *fakefs) entryForName(name string) *fakeEntry {
+func (fs *fakeFS) entryForName(name string) *fakeEntry {
 	// bug: lookup doesn't work through symlinks.
 	if fs.insens {
-		name = UnicodeLowercase(name)
+		name = UnicodeLowercaseNormalized(name)
 	}
 
 	name = filepath.ToSlash(name)
@@ -210,7 +210,7 @@ func (fs *fakefs) entryForName(name string) *fakeEntry {
 	return entry
 }
 
-func (fs *fakefs) Chmod(name string, mode FileMode) error {
+func (fs *fakeFS) Chmod(name string, mode FileMode) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Chmod++
@@ -223,7 +223,7 @@ func (fs *fakefs) Chmod(name string, mode FileMode) error {
 	return nil
 }
 
-func (fs *fakefs) Lchown(name string, uid, gid int) error {
+func (fs *fakeFS) Lchown(name string, uid, gid int) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Lchown++
@@ -237,7 +237,7 @@ func (fs *fakefs) Lchown(name string, uid, gid int) error {
 	return nil
 }
 
-func (fs *fakefs) Chtimes(name string, atime time.Time, mtime time.Time) error {
+func (fs *fakeFS) Chtimes(name string, atime time.Time, mtime time.Time) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Chtimes++
@@ -250,7 +250,7 @@ func (fs *fakefs) Chtimes(name string, atime time.Time, mtime time.Time) error {
 	return nil
 }
 
-func (fs *fakefs) create(name string) (*fakeEntry, error) {
+func (fs *fakeFS) create(name string) (*fakeEntry, error) {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Create++
@@ -285,7 +285,7 @@ func (fs *fakefs) create(name string) (*fakeEntry, error) {
 	}
 
 	if fs.insens {
-		base = UnicodeLowercase(base)
+		base = UnicodeLowercaseNormalized(base)
 	}
 
 	if fs.withContent {
@@ -296,7 +296,7 @@ func (fs *fakefs) create(name string) (*fakeEntry, error) {
 	return new, nil
 }
 
-func (fs *fakefs) Create(name string) (File, error) {
+func (fs *fakeFS) Create(name string) (File, error) {
 	entry, err := fs.create(name)
 	if err != nil {
 		return nil, err
@@ -307,7 +307,7 @@ func (fs *fakefs) Create(name string) (File, error) {
 	return &fakeFile{fakeEntry: entry}, nil
 }
 
-func (fs *fakefs) CreateSymlink(target, name string) error {
+func (fs *fakeFS) CreateSymlink(target, name string) error {
 	entry, err := fs.create(name)
 	if err != nil {
 		return err
@@ -317,7 +317,7 @@ func (fs *fakefs) CreateSymlink(target, name string) error {
 	return nil
 }
 
-func (fs *fakefs) DirNames(name string) ([]string, error) {
+func (fs *fakeFS) DirNames(name string) ([]string, error) {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.DirNames++
@@ -336,7 +336,7 @@ func (fs *fakefs) DirNames(name string) ([]string, error) {
 	return names, nil
 }
 
-func (fs *fakefs) Lstat(name string) (FileInfo, error) {
+func (fs *fakeFS) Lstat(name string) (FileInfo, error) {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Lstat++
@@ -355,7 +355,7 @@ func (fs *fakefs) Lstat(name string) (FileInfo, error) {
 	return info, nil
 }
 
-func (fs *fakefs) Mkdir(name string, perm FileMode) error {
+func (fs *fakeFS) Mkdir(name string, perm FileMode) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Mkdir++
@@ -373,7 +373,7 @@ func (fs *fakefs) Mkdir(name string, perm FileMode) error {
 		return os.ErrExist
 	}
 	if fs.insens {
-		key = UnicodeLowercase(key)
+		key = UnicodeLowercaseNormalized(key)
 	}
 	if _, ok := entry.children[key]; ok {
 		return os.ErrExist
@@ -389,7 +389,7 @@ func (fs *fakefs) Mkdir(name string, perm FileMode) error {
 	return nil
 }
 
-func (fs *fakefs) MkdirAll(name string, perm FileMode) error {
+func (fs *fakeFS) MkdirAll(name string, perm FileMode) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.MkdirAll++
@@ -402,7 +402,7 @@ func (fs *fakefs) MkdirAll(name string, perm FileMode) error {
 	for _, comp := range comps {
 		key := comp
 		if fs.insens {
-			key = UnicodeLowercase(key)
+			key = UnicodeLowercaseNormalized(key)
 		}
 
 		next, ok := entry.children[key]
@@ -426,7 +426,7 @@ func (fs *fakefs) MkdirAll(name string, perm FileMode) error {
 	return nil
 }
 
-func (fs *fakefs) Open(name string) (File, error) {
+func (fs *fakeFS) Open(name string) (File, error) {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Open++
@@ -443,7 +443,7 @@ func (fs *fakefs) Open(name string) (File, error) {
 	return &fakeFile{fakeEntry: entry}, nil
 }
 
-func (fs *fakefs) OpenFile(name string, flags int, mode FileMode) (File, error) {
+func (fs *fakeFS) OpenFile(name string, flags int, mode FileMode) (File, error) {
 	if flags&os.O_CREATE == 0 {
 		return fs.Open(name)
 	}
@@ -465,7 +465,7 @@ func (fs *fakefs) OpenFile(name string, flags int, mode FileMode) (File, error)
 	}
 
 	if fs.insens {
-		key = UnicodeLowercase(key)
+		key = UnicodeLowercaseNormalized(key)
 	}
 	if flags&os.O_EXCL != 0 {
 		if _, ok := entry.children[key]; ok {
@@ -486,7 +486,7 @@ func (fs *fakefs) OpenFile(name string, flags int, mode FileMode) (File, error)
 	return &fakeFile{fakeEntry: newEntry}, nil
 }
 
-func (fs *fakefs) ReadSymlink(name string) (string, error) {
+func (fs *fakeFS) ReadSymlink(name string) (string, error) {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.ReadSymlink++
@@ -501,14 +501,14 @@ func (fs *fakefs) ReadSymlink(name string) (string, error) {
 	return entry.dest, nil
 }
 
-func (fs *fakefs) Remove(name string) error {
+func (fs *fakeFS) Remove(name string) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Remove++
 	time.Sleep(fs.latency)
 
 	if fs.insens {
-		name = UnicodeLowercase(name)
+		name = UnicodeLowercaseNormalized(name)
 	}
 
 	entry := fs.entryForName(name)
@@ -524,19 +524,19 @@ func (fs *fakefs) Remove(name string) error {
 	return nil
 }
 
-func (fs *fakefs) RemoveAll(name string) error {
+func (fs *fakeFS) RemoveAll(name string) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.RemoveAll++
 	time.Sleep(fs.latency)
 
 	if fs.insens {
-		name = UnicodeLowercase(name)
+		name = UnicodeLowercaseNormalized(name)
 	}
 
 	entry := fs.entryForName(filepath.Dir(name))
 	if entry == nil {
-		return nil // all tested real systems exibit this behaviour
+		return nil // all tested real systems exhibit this behaviour
 	}
 
 	// RemoveAll is easy when the file system uses garbage collection under
@@ -545,7 +545,7 @@ func (fs *fakefs) RemoveAll(name string) error {
 	return nil
 }
 
-func (fs *fakefs) Rename(oldname, newname string) error {
+func (fs *fakeFS) Rename(oldname, newname string) error {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	fs.counters.Rename++
@@ -555,8 +555,8 @@ func (fs *fakefs) Rename(oldname, newname string) error {
 	newKey := filepath.Base(newname)
 
 	if fs.insens {
-		oldKey = UnicodeLowercase(oldKey)
-		newKey = UnicodeLowercase(newKey)
+		oldKey = UnicodeLowercaseNormalized(oldKey)
+		newKey = UnicodeLowercaseNormalized(newKey)
 	}
 
 	p0 := fs.entryForName(filepath.Dir(oldname))
@@ -595,59 +595,63 @@ func (fs *fakefs) Rename(oldname, newname string) error {
 	return nil
 }
 
-func (fs *fakefs) Stat(name string) (FileInfo, error) {
+func (fs *fakeFS) Stat(name string) (FileInfo, error) {
 	return fs.Lstat(name)
 }
 
-func (fs *fakefs) SymlinksSupported() bool {
+func (fs *fakeFS) SymlinksSupported() bool {
 	return false
 }
 
-func (fs *fakefs) Walk(name string, walkFn WalkFunc) error {
+func (fs *fakeFS) Walk(name string, walkFn WalkFunc) error {
 	return errors.New("not implemented")
 }
 
-func (fs *fakefs) Watch(path string, ignore Matcher, ctx context.Context, ignorePerms bool) (<-chan Event, <-chan error, error) {
+func (fs *fakeFS) Watch(path string, ignore Matcher, ctx context.Context, ignorePerms bool) (<-chan Event, <-chan error, error) {
 	return nil, nil, ErrWatchNotSupported
 }
 
-func (fs *fakefs) Hide(name string) error {
+func (fs *fakeFS) Hide(name string) error {
 	return nil
 }
 
-func (fs *fakefs) Unhide(name string) error {
+func (fs *fakeFS) Unhide(name string) error {
 	return nil
 }
 
-func (fs *fakefs) Glob(pattern string) ([]string, error) {
+func (fs *fakeFS) Glob(pattern string) ([]string, error) {
 	// gnnh we don't seem to actually require this in practice
 	return nil, errors.New("not implemented")
 }
 
-func (fs *fakefs) Roots() ([]string, error) {
+func (fs *fakeFS) Roots() ([]string, error) {
 	return []string{"/"}, nil
 }
 
-func (fs *fakefs) Usage(name string) (Usage, error) {
+func (fs *fakeFS) Usage(name string) (Usage, error) {
 	return Usage{}, errors.New("not implemented")
 }
 
-func (fs *fakefs) Type() FilesystemType {
+func (fs *fakeFS) Type() FilesystemType {
 	return FilesystemTypeFake
 }
 
-func (fs *fakefs) URI() string {
+func (fs *fakeFS) URI() string {
 	return fs.uri
 }
 
-func (fs *fakefs) SameFile(fi1, fi2 FileInfo) bool {
+func (fs *fakeFS) Options() []Option {
+	return nil
+}
+
+func (fs *fakeFS) SameFile(fi1, fi2 FileInfo) bool {
 	// BUG: real systems base file sameness on path, inodes, etc
 	// we try our best, but FileInfo just doesn't have enough data
 	// so there be false positives, especially on Windows
 	// where ModTime is not that precise
 	var ok bool
 	if fs.insens {
-		ok = UnicodeLowercase(fi1.Name()) == UnicodeLowercase(fi2.Name())
+		ok = UnicodeLowercaseNormalized(fi1.Name()) == UnicodeLowercaseNormalized(fi2.Name())
 	} else {
 		ok = fi1.Name() == fi2.Name()
 	}
@@ -655,17 +659,25 @@ func (fs *fakefs) SameFile(fi1, fi2 FileInfo) bool {
 	return ok && fi1.ModTime().Equal(fi2.ModTime()) && fi1.Mode() == fi2.Mode() && fi1.IsDir() == fi2.IsDir() && fi1.IsRegular() == fi2.IsRegular() && fi1.IsSymlink() == fi2.IsSymlink() && fi1.Owner() == fi2.Owner() && fi1.Group() == fi2.Group()
 }
 
-func (fs *fakefs) resetCounters() {
+func (fs *fakeFS) underlying() (Filesystem, bool) {
+	return nil, false
+}
+
+func (fs *fakeFS) wrapperType() filesystemWrapperType {
+	return filesystemWrapperTypeNone
+}
+
+func (fs *fakeFS) resetCounters() {
 	fs.mut.Lock()
-	fs.counters = fakefsCounters{}
+	fs.counters = fakeFSCounters{}
 	fs.mut.Unlock()
 }
 
-func (fs *fakefs) reportMetricsPerOp(b *testing.B) {
+func (fs *fakeFS) reportMetricsPerOp(b *testing.B) {
 	fs.reportMetricsPer(b, 1, "op")
 }
 
-func (fs *fakefs) reportMetricsPer(b *testing.B, divisor float64, unit string) {
+func (fs *fakeFS) reportMetricsPer(b *testing.B, divisor float64, unit string) {
 	fs.mut.Lock()
 	defer fs.mut.Unlock()
 	b.ReportMetric(float64(fs.counters.Lstat)/divisor/float64(b.N), "Lstat/"+unit)
@@ -873,7 +885,13 @@ func (f *fakeFile) Truncate(size int64) error {
 	defer f.mut.Unlock()
 
 	if f.content != nil {
-		f.content = f.content[:int(size)]
+		if int64(cap(f.content)) < size {
+			c := make([]byte, size)
+			copy(c[:len(f.content)], f.content)
+			f.content = c
+		} else {
+			f.content = f.content[:int(size)]
+		}
 	}
 	f.rng = nil
 	f.size = size
